"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __await = (this && this.__await) || function (v) { return this instanceof __await ? (this.v = v, this) : new __await(v); }
var __asyncValues = (this && this.__asyncValues) || function (o) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var m = o[Symbol.asyncIterator];
    return m ? m.call(o) : typeof __values === "function" ? __values(o) : o[Symbol.iterator]();
};
var __asyncGenerator = (this && this.__asyncGenerator) || function (thisArg, _arguments, generator) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var g = generator.apply(thisArg, _arguments || []), i, q = [];
    return i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i;
    function verb(n) { if (g[n]) i[n] = function (v) { return new Promise(function (a, b) { q.push([n, v, a, b]) > 1 || resume(n, v); }); }; }
    function resume(n, v) { try { step(g[n](v)); } catch (e) { settle(q[0][3], e); } }
    function step(r) { r.value instanceof __await ? Promise.resolve(r.value.v).then(fulfill, reject) : settle(q[0][2], r);  }
    function fulfill(value) { resume("next", value); }
    function reject(value) { resume("throw", value); }
    function settle(f, v) { if (f(v), q.shift(), q.length) resume(q[0][0], q[0][1]); }
};
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const child_process = require("child_process");
const util = require("util");
const lazily_async_1 = require("lazily-async");
const exception_1 = require("./exception");
const import_module_1 = require("./import-module");
const exec = util.promisify(child_process.exec);
// prettier-ignore
/* Options: */
const options = [
    "-a",
    "-c",
    "-e",
    "-f",
    "-g",
    "-i",
    "-j",
    "-l",
    "-m",
    "-n",
    "-q",
    "-p",
    "-r",
    "-s",
    "-t",
    "-w",
    "--error",
];
class QuotedExpression {
    constructor(str) {
        this.str = str;
    }
}
class PipelineItem {
    constructor(previousItem, name) {
        this.previousItem = previousItem;
        this.name = name;
    }
}
exports.PipelineItem = PipelineItem;
class PipelineValue extends PipelineItem {
    constructor(value, previousItem, name) {
        super(previousItem, name);
        this.value = value;
    }
    clone(name) {
        return new PipelineValue(this.value, this.previousItem, name);
    }
}
exports.PipelineValue = PipelineValue;
class PipelineError extends PipelineItem {
    constructor(message, error, previousItem, name) {
        super(previousItem, name);
        this.message = message;
        this.error = error;
    }
    clone(name) {
        return new PipelineError(this.message, this.error, this.previousItem, name);
    }
}
exports.PipelineError = PipelineError;
class EvalError {
    constructor(error) {
        this.error = error;
    }
}
function evalWithCatch(exp) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            const fn = yield eval(exp);
            return function () {
                return __awaiter(this, arguments, void 0, function* () {
                    try {
                        return yield fn.apply(undefined, arguments);
                    }
                    catch (ex) {
                        return new EvalError(ex);
                    }
                });
            };
        }
        catch (ex) {
            return () => {
                return new EvalError(ex);
            };
        }
    });
}
function shellEscape(str) {
    return str
        .replace(/([^A-Za-z0-9_\-.,:\/@\n])/g, "\\$1")
        .replace(/\n/g, "'\n'");
}
function evalExpression(exp, input, nextArgs, isInitialInput) {
    return __awaiter(this, void 0, void 0, function* () {
        return isInitialInput
            ? yield (() => __awaiter(this, void 0, void 0, function* () {
                const code = `async () => (${exp})`;
                const fn = yield evalWithCatch(code);
                const input = yield fn();
                return input instanceof EvalError
                    ? lazily_async_1.Seq.of([
                        new PipelineError(`Failed to evaluate expression: ${exp}.`, input.error)
                    ])
                    : Array.isArray(input)
                        ? lazily_async_1.Seq.of(input.map(i => new PipelineValue(i)))
                        : lazily_async_1.Seq.of([new PipelineValue(input)]);
            }))()
            : yield (() => __awaiter(this, void 0, void 0, function* () {
                const code = `async (x, i) => (${exp})`;
                return input.map((x, i) => __awaiter(this, void 0, void 0, function* () {
                    return x instanceof PipelineError
                        ? x
                        : x instanceof PipelineValue
                            ? yield (() => __awaiter(this, void 0, void 0, function* () {
                                const fn = yield evalWithCatch(code);
                                const result = yield fn(yield x.value, i);
                                return result instanceof EvalError
                                    ? new PipelineError(`Failed to evaluate expression: ${exp}.`, result.error, x)
                                    : new PipelineValue(result, x);
                            }))()
                            : exception_1.default(`Invalid item ${x} in pipeline.`);
                }));
            }))();
    });
}
function shellCmd(template, input, nextArgs, isInitialInput) {
    return __awaiter(this, void 0, void 0, function* () {
        const fn = yield evalWithCatch(`async (x, i) => \`${template}\``);
        return isInitialInput
            ? yield (() => __awaiter(this, void 0, void 0, function* () {
                try {
                    const cmd = yield fn();
                    return cmd instanceof EvalError
                        ? lazily_async_1.Seq.of([
                            new PipelineError(`Failed to execute shell command: ${template}`, cmd.error)
                        ])
                        : yield (() => __awaiter(this, void 0, void 0, function* () {
                            const { stdout } = yield exec(cmd);
                            return lazily_async_1.Seq.of(stdout
                                .split("\n")
                                .filter(x => x !== "")
                                .map(x => x.replace(/\n$/, ""))
                                .map(i => new PipelineValue(i)));
                        }))();
                }
                catch (ex) {
                    return lazily_async_1.Seq.of([
                        new PipelineError(`Failed to execute shell command: ${template}`, ex)
                    ]);
                }
            }))()
            : yield (() => __awaiter(this, void 0, void 0, function* () {
                return input.map((x, i) => __awaiter(this, void 0, void 0, function* () {
                    return x instanceof PipelineError
                        ? x
                        : x instanceof PipelineValue
                            ? yield (() => __awaiter(this, void 0, void 0, function* () {
                                try {
                                    const value = yield x.value;
                                    const cmd = yield fn(typeof value === "string" ? shellEscape(value) : value, i);
                                    const { stdout } = yield exec(cmd);
                                    const items = stdout
                                        .split("\n")
                                        .filter(x => x !== "")
                                        .map(x => x.replace(/\n$/, ""));
                                    return new PipelineValue(items.length === 1 ? items[0] : items, x);
                                }
                                catch (ex) {
                                    return new PipelineError(`Failed to execute shell command: ${template}`, ex);
                                }
                            }))()
                            : exception_1.default(`Invalid item ${x} in pipeline.`);
                }));
            }))();
    });
}
function evalImport(filename, alias) {
    return __awaiter(this, void 0, void 0, function* () {
        const isRelative = filename.startsWith("./") ||
            filename.startsWith("../") ||
            filename.endsWith(".js");
        const filePath = isRelative ? path.join(process.cwd(), filename) : filename;
        yield import_module_1.default(filePath, alias, isRelative);
    });
}
function filter(exp, input) {
    return __awaiter(this, void 0, void 0, function* () {
        const code = `async (x, i) => (${exp})`;
        const fn = yield evalWithCatch(code);
        return input.filter((x, i) => __awaiter(this, void 0, void 0, function* () {
            return x instanceof PipelineError
                ? true
                : x instanceof PipelineValue
                    ? yield (() => __awaiter(this, void 0, void 0, function* () {
                        const result = yield fn(yield x.value, i);
                        return result instanceof EvalError ? true : result;
                    }))()
                    : exception_1.default(`Invalid item ${x} in pipeline.`);
        }));
    });
}
function flatMap(exp, input) {
    return __awaiter(this, void 0, void 0, function* () {
        const code = `async (x, i) => (${exp})`;
        const fn = yield evalWithCatch(code);
        return input.flatMap((x, i) => __awaiter(this, void 0, void 0, function* () {
            return x instanceof PipelineError
                ? [x]
                : x instanceof PipelineValue
                    ? yield (() => __awaiter(this, void 0, void 0, function* () {
                        const result = yield fn(yield x.value, i);
                        return result instanceof EvalError
                            ? new PipelineError(`Failed to evaluate expression: ${exp}.`, result.error, x)
                            : result.map((r) => new PipelineValue(r, x));
                    }))()
                    : exception_1.default(`Invalid item ${x} in pipeline.`);
        }));
    });
}
function reduce(exp, input, initialValueExp) {
    return __awaiter(this, void 0, void 0, function* () {
        const code = `async (acc, x, i) => (${exp})`;
        const fn = yield evalWithCatch(code);
        const initialValueCode = `async () => (${initialValueExp})`;
        const getInitialValue = yield evalWithCatch(initialValueCode);
        const initialValue = yield getInitialValue();
        const output = initialValue instanceof EvalError
            ? new PipelineError(`Failed to evaluate expression: ${initialValue}.`, initialValue.error)
            : yield input.reduce((acc, x, i) => __awaiter(this, void 0, void 0, function* () {
                return acc instanceof PipelineError
                    ? acc
                    : x instanceof PipelineError
                        ? x
                        : x instanceof PipelineValue
                            ? yield (() => __awaiter(this, void 0, void 0, function* () {
                                const result = yield fn(acc, yield x.value, i);
                                return result instanceof EvalError
                                    ? new PipelineError(`Failed to evaluate expression: ${exp}.`, result.error, x)
                                    : result;
                            }))()
                            : exception_1.default(`Invalid item ${x} in pipeline.`);
            }), initialValue);
        return output instanceof PipelineError ? output : new PipelineValue(output);
    });
}
/* Consume parameters until we reach an option flag (-p, -e etc) */
function munch(parts, numOtherExpressions = 0) {
    function doMunch(parts, args, cursor) {
        return !parts.length || options.includes(parts[0])
            ? { cursor, args }
            : doMunch(parts.slice(1), args.concat(parts[0]), cursor + 1);
    }
    const isQuoted = parts[0] === "-q";
    const { cursor, args } = doMunch(parts.slice(isQuoted ? 1 : 0), [], isQuoted ? 1 : 0);
    const result = numOtherExpressions > 0
        ? {
            cursor,
            expression: `${args.slice(0, -numOtherExpressions).join(" ")}`,
            otherExpressions: args.slice(-numOtherExpressions)
        }
        : { cursor, expression: `${args.join(" ")}` };
    return Object.assign({}, result, { expression: isQuoted ? `"${result.expression}"` : result.expression });
}
function findNamedValue(name, pipelineItem) {
    return typeof pipelineItem !== "undefined"
        ? pipelineItem.name === name
            ? pipelineItem
            : findNamedValue(name, pipelineItem.previousItem)
        : undefined;
}
function evaluate(args, pipedValues = [], mustPrint = true, onLog = () => { }, onWrite = () => { }) {
    return __awaiter(this, void 0, void 0, function* () {
        return yield evaluateInternal(args, [], lazily_async_1.Seq.of(pipedValues.map(x => new PipelineValue(x))), mustPrint, onLog, onWrite, pipedValues.length === 0, []);
    });
}
exports.evaluate = evaluate;
function evaluateInternal(args, prevArgs, input, mustPrint, onLog, onWrite, isInitialInput, expressionStack) {
    return __awaiter(this, void 0, void 0, function* () {
        const cases = [
            /* Enumerate sequence into an array */
            [
                x => x === "-a",
                () => __awaiter(this, void 0, void 0, function* () {
                    const items = yield input.toArray();
                    return yield evalShorthand(args.slice(1), args, lazily_async_1.Seq.of([
                        new PipelineValue(items.map(x => (x instanceof PipelineValue ? x.value : x)))
                    ]), false);
                })
            ],
            /* Combine multiple named streams */
            [
                x => x === "-c",
                () => __awaiter(this, void 0, void 0, function* () {
                    return yield evalShorthand(args.slice(2), args, input.map(x => {
                        const streams = args[1].split(",");
                        return new PipelineValue(streams.map(name => {
                            const item = findNamedValue(name, x);
                            return item instanceof PipelineValue ? item.value : item;
                        }), x);
                    }), false);
                })
            ],
            /* Execute shell command */
            [
                x => x === "-e",
                () => __awaiter(this, void 0, void 0, function* () {
                    const { cursor, expression } = munch(args.slice(1));
                    return yield evalShorthand(args.slice(cursor + 1), args, yield shellCmd(expression, input, args.slice(cursor + 1), isInitialInput), false);
                })
            ],
            /* Error handling */
            [
                x => x === "--error",
                () => __awaiter(this, void 0, void 0, function* () {
                    const { cursor, expression } = munch(args.slice(1));
                    const newSeq = input.map((x, i) => __awaiter(this, void 0, void 0, function* () {
                        const fn = yield evalWithCatch(`async (x, i) => (${expression})`);
                        return x instanceof PipelineError
                            ? yield (() => __awaiter(this, void 0, void 0, function* () {
                                const result = yield fn(x, i);
                                return result instanceof EvalError
                                    ? new PipelineError(`Failed to evaluate error expression: ${expression}.`, result.error, x)
                                    : new PipelineValue(result, x);
                            }))()
                            : x;
                    }));
                    return yield evalShorthand(args.slice(cursor + 1), args, newSeq, false);
                })
            ],
            /* Filter */
            [
                x => x === "-f",
                () => __awaiter(this, void 0, void 0, function* () {
                    const { cursor, expression } = munch(args.slice(1));
                    const filtered = yield filter(expression, input);
                    return yield evalShorthand(args.slice(cursor + 1), args, filtered, false);
                })
            ],
            /* Recurse/Goto */
            [
                x => x === "-g",
                () => __awaiter(this, void 0, void 0, function* () {
                    const name = args[1];
                    const { cursor, expression } = munch(args.slice(2));
                    const recursePoint = expressionStack.find(e => e.name === name);
                    const fn = yield evalWithCatch(`(x, i) => (${expression})`);
                    const newSeq = input.map((x, i) => __awaiter(this, void 0, void 0, function* () {
                        return recursePoint
                            ? x instanceof PipelineValue
                                ? yield (() => __awaiter(this, void 0, void 0, function* () {
                                    const predicateResult = yield fn(yield x.value, i);
                                    return predicateResult === true
                                        ? yield (() => __awaiter(this, void 0, void 0, function* () {
                                            const recurseArgs = recursePoint.args.slice(0, recursePoint.args.length -
                                                (args.length - (cursor + 2)));
                                            const innerEvalResult = yield evaluateInternal(recurseArgs, [], lazily_async_1.Seq.of([x]), mustPrint, onLog, onWrite, false, []);
                                            const results = yield innerEvalResult.result.toArray();
                                            const result = results[0];
                                            return result instanceof PipelineValue
                                                ? new PipelineValue(result.value, x)
                                                : result;
                                        }))()
                                        : x;
                                }))()
                                : x
                            : new PipelineError(`The expression ${name} was not found.`, new Error(`Missing expression ${name}.`));
                    }));
                    return yield evalShorthand(args.slice(cursor + 2), args, newSeq, false);
                })
            ],
            /* Named Export */
            [
                x => x === "-i",
                () => __awaiter(this, void 0, void 0, function* () {
                    yield evalImport(args[1], args[2]);
                    return yield evaluateInternal(args.slice(3), args, input, mustPrint, onLog, onWrite, isInitialInput, expressionStack);
                })
            ],
            /* JS expressions */
            [
                x => x === "-j",
                () => __awaiter(this, void 0, void 0, function* () {
                    const { cursor, expression } = munch(args.slice(1));
                    return yield evalShorthand(args.slice(cursor + 1), args, yield evalExpression(expression, input, args.slice(cursor + 1), isInitialInput), false);
                })
            ],
            /* Logging */
            [x => x === "-l", getPrinter(onLog)(input, args)],
            /* Flatmap */
            [
                x => x === "-m",
                () => __awaiter(this, void 0, void 0, function* () {
                    const { cursor, expression } = munch(args.slice(1));
                    const mapped = yield flatMap(expression, input);
                    return yield evalShorthand(args.slice(cursor + 1), args, mapped, false);
                })
            ],
            /* Named Expressions */
            [
                x => x === "-n",
                () => __awaiter(this, void 0, void 0, function* () {
                    const newSeq = input.map((x, i) => __awaiter(this, void 0, void 0, function* () { return x.clone(args[1]); }));
                    return yield evaluateInternal(args.slice(2), args, newSeq, mustPrint, onLog, onWrite, false, expressionStack.concat({ name: args[1], args: prevArgs }));
                })
            ],
            /* Error handling. Handled by shell, ignore */
            [
                x => x === "--ignoreerror" || x === "--printerror",
                () => __awaiter(this, void 0, void 0, function* () {
                    return yield evaluateInternal(args.slice(1), args, input, mustPrint, onLog, onWrite, isInitialInput, expressionStack);
                })
            ],
            /* Print */
            [
                x => x === "-p",
                () => __awaiter(this, void 0, void 0, function* () {
                    return yield evaluateInternal(args.slice(1), args, input, false, onLog, onWrite, isInitialInput, expressionStack);
                })
            ],
            /* Reduce */
            [
                x => x === "-r",
                () => __awaiter(this, void 0, void 0, function* () {
                    const { cursor, expression, otherExpressions } = munch(args.slice(1), 1);
                    return typeof otherExpressions !== "undefined"
                        ? yield (() => __awaiter(this, void 0, void 0, function* () {
                            const initialValue = otherExpressions[0];
                            return yield (() => __awaiter(this, void 0, void 0, function* () {
                                const reduced = yield reduce(expression, input, initialValue);
                                return yield evalShorthand(args.slice(cursor + 1), args, lazily_async_1.Seq.of([reduced]), false);
                            }))();
                        }))()
                        : exception_1.default(`Failed to evaluate initial value expression.`);
                })
            ],
            /* Seek a named result */
            [
                x => x === "-s",
                () => __awaiter(this, void 0, void 0, function* () {
                    return yield evalShorthand(args.slice(2), args, input.map(x => {
                        return new PipelineValue((() => {
                            const item = findNamedValue(args[1], x);
                            return item instanceof PipelineValue ? item.value : item;
                        })(), x);
                    }), false);
                })
            ],
            /* Terminate the pipeline */
            [
                x => x === "-t",
                () => __awaiter(this, void 0, void 0, function* () {
                    const { cursor, expression } = munch(args.slice(1));
                    function asyncGenerator() {
                        return __asyncGenerator(this, arguments, function* asyncGenerator_1() {
                            const fn = yield __await(evalWithCatch(`(x, i) => (${expression})`));
                            let i = 0;
                            try {
                                for (var input_1 = __asyncValues(input), input_1_1; input_1_1 = yield __await(input_1.next()), !input_1_1.done;) {
                                    const x = yield __await(input_1_1.value);
                                    if (x instanceof PipelineValue) {
                                        const result = yield __await(fn(yield __await(x.value), i));
                                        if (result instanceof EvalError) {
                                            return new PipelineError(`Failed to evaluate expression: ${expression}.`, result.error, x);
                                        }
                                        if (result === true) {
                                            return x;
                                        }
                                        yield x;
                                    }
                                    else {
                                        yield x;
                                    }
                                    i++;
                                }
                            }
                            catch (e_1_1) { e_1 = { error: e_1_1 }; }
                            finally {
                                try {
                                    if (input_1_1 && !input_1_1.done && (_a = input_1.return)) yield __await(_a.call(input_1));
                                }
                                finally { if (e_1) throw e_1.error; }
                            }
                            var e_1, _a;
                        });
                    }
                    return yield evalShorthand(args.slice(cursor + 1), args, new lazily_async_1.Seq(asyncGenerator), false);
                })
            ],
            /* Writing */
            [x => x === "-w", getPrinter(onWrite)(input, args)],
            /* Everything else as JS expressions */
            [
                x => true,
                () => __awaiter(this, void 0, void 0, function* () {
                    const { cursor, expression } = munch(args);
                    return yield evalShorthand(args.slice(cursor), args, yield evalExpression(expression, input, args.slice(cursor), isInitialInput), false);
                })
            ]
        ];
        function getPrinter(printFn) {
            return (input, args) => () => __awaiter(this, void 0, void 0, function* () {
                const { cursor, expression } = munch(args.slice(1));
                const fn = yield evalWithCatch(`(x, i) => (${expression})`);
                const newSeq = input.map((x, i) => __awaiter(this, void 0, void 0, function* () {
                    if (x instanceof PipelineValue) {
                        const result = yield fn(yield x.value, i);
                        printFn(result instanceof EvalError
                            ? `Failed to evaluate expression: ${expression}.`
                            : result);
                    }
                    return x;
                }));
                return yield evalShorthand(args.slice(cursor + 1), args, newSeq, isInitialInput);
            });
        }
        function evalShorthand(args, prevArgs, input, isInitialInput) {
            return __awaiter(this, void 0, void 0, function* () {
                return yield evaluateInternal(args, prevArgs, input, mustPrint, onLog, onWrite, isInitialInput, expressionStack);
            });
        }
        return args.length
            ? yield (() => __awaiter(this, void 0, void 0, function* () {
                const handler = cases.find(([predicate]) => predicate(args[0]));
                return handler
                    ? yield handler[1]()
                    : exception_1.default(`Cannot parse option ${args[0]}.`);
            }))()
            : { mustPrint, result: input };
    });
}
//# sourceMappingURL=basho-eval.js.map